from nltk import downloader
from nltk.corpus import wordnet

from chia.knowledge.relation import RelationSource


class WordNetAccess(RelationSource):
    def __init__(self):
        # Check if required corpora are there
        corpus_downloader = downloader.Downloader()
        for corpus in ("wordnet", "omw"):
            if not corpus_downloader.is_installed(corpus):
                corpus_downloader.download(corpus)

        try:
            wordnet.synset("dog.n.01")
        except LookupError:
            raise ValueError("Could not perform test lookup!")

    def _get_hypernyms(self, synset):
        return {
            f"WordNet3.0::{hsynset.name()}"
            for hsynset in wordnet.synset(synset).hypernyms()
        }

    def _get_hyponyms(self, synset):
        return {
            f"WordNet3.0::{hsynset.name()}"
            for hsynset in wordnet.synset(synset).hyponyms()
        }

    def get_right_for(self, uid_left):
        if uid_left.startswith("WordNet3.0::"):
            return self._get_hypernyms(uid_left[12:])
        else:
            return set()

    def get_left_for(self, uid_right):
        if uid_right.startswith("WordNet3.0::"):
            return self._get_hyponyms(uid_right[12:])
        else:
            return set()



import multiprocessing
import pickle as pkl
import time

import numpy as np

from chia import components, instrumentation
from chia.components.base_models.incremental_model import ProbabilityOutputModel
from chia.components.base_models.keras import (
    keras_featureextractor,
    keras_preprocessor,
    keras_trainer,
)
from chia.components.classifiers import keras_hierarchicalclassification
from chia.helpers import batches_from, make_generator_faster


class KerasBaseModel(ProbabilityOutputModel, instrumentation.Observable):
    def __init__(
        self,
        classifier: keras_hierarchicalclassification.KerasHierarchicalClassifier,
        feature_extractor: keras_featureextractor.KerasFeatureExtractor,
        preprocessor: keras_preprocessor.KerasPreprocessor,
        trainer: keras_trainer.KerasTrainer,
        batch_size,
    ):
        instrumentation.Observable.__init__(self)

        self.classifier = classifier

        # Batch size
        self.batch_size = batch_size

        # Feature Extractor
        self.feature_extractor_new = feature_extractor

        # Preprocessing
        self.preprocessor = preprocessor

        # Trainer
        self.trainer = trainer

    def observe_inner(self, samples, gt_resource_id, progress_callback=None):
        self.trainer.observe_inner(samples, gt_resource_id, progress_callback)

    def observe(self, samples, gt_resource_id, progress_callback=None):
        self.classifier.observe(samples, gt_resource_id)
        self.observe_inner(samples, gt_resource_id, progress_callback)

    def predict(self, samples, prediction_resource_id):
        return_samples = []
        batch_size = self.batch_size

        total_time_data = 0.0
        total_time_preprocess = 0.0
        total_time_features = 0.0
        total_time_cls = 0.0
        total_time_write = 0.0

        def my_gen():
            pool = multiprocessing.pool.ThreadPool(4)
            for small_batch_ in batches_from(samples, batch_size=batch_size):
                built_image_batch_ = self.build_image_batch(small_batch_, pool)
                yield small_batch_, built_image_batch_
            pool.close()

        tp_before_data = time.time()
        faster_generator = make_generator_faster(
            my_gen, method="threading", observable=self, max_buffer_size=50
        )
        for (small_batch, built_image_batch) in faster_generator:
            tp_before_preprocess = time.time()
            image_batch = self.preprocessor.preprocess_image_batch(
                built_image_batch, is_training=False
            )
            tp_before_features = time.time()
            feature_batch = self.feature_extractor_new(image_batch, training=False)
            tp_before_cls = time.time()
            predictions_dist = self.classifier.predict_dist(feature_batch)
            predictions = [
                sorted(prediction_dist, key=lambda x: x[1], reverse=True)[0][0]
                for prediction_dist in predictions_dist
            ]

            tp_before_write = time.time()
            return_samples += [
                sample.add_resource(
                    self.__class__.__name__, prediction_resource_id, prediction
                ).add_resource(
                    self.__class__.__name__,
                    prediction_resource_id + "_dist",
                    prediction_dist,
                )
                for prediction, prediction_dist, sample in zip(
                    predictions, predictions_dist, small_batch
                )
            ]
            tp_loop_done = time.time()
            total_time_data += tp_before_preprocess - tp_before_data
            total_time_preprocess += tp_before_features - tp_before_preprocess
            total_time_features += tp_before_cls - tp_before_features
            total_time_cls += tp_before_write - tp_before_cls
            total_time_write += tp_loop_done - tp_before_write

            tp_before_data = time.time()

        self.log_debug("Predict done.")
        self.log_debug(f"Time (data): {total_time_data}")
        self.log_debug(f"Time (preprocess): {total_time_preprocess}")
        self.log_debug(f"Time (features): {total_time_features}")
        self.log_debug(f"Time (cls): {total_time_cls}")
        self.log_debug(f"Time (write): {total_time_write}")
        total_time_overall = (
            total_time_data
            + total_time_preprocess
            + total_time_features
            + total_time_cls
            + total_time_write
        )
        self.log_debug(f"Total time: {total_time_overall}")
        return return_samples

    def predict_probabilities(self, samples, prediction_dist_resource_id):
        return_samples = []
        batch_size = self.batch_size

        def my_gen():
            pool = multiprocessing.pool.ThreadPool(4)
            for small_batch_ in batches_from(samples, batch_size=batch_size):
                built_image_batch_ = self.build_image_batch(small_batch_, pool)
                yield small_batch_, built_image_batch_

        for small_batch, built_image_batch in make_generator_faster(
            my_gen, method="synchronous", observable=self, max_buffer_size=5
        ):
            image_batch = self.preprocessor.preprocess_image_batch(
                built_image_batch, is_training=False
            )
            feature_batch = self.feature_extractor_new(image_batch, training=False)
            predictions = self.classifier.predict_dist(feature_batch)
            return_samples += [
                sample.add_resource(
                    self.__class__.__name__, prediction_dist_resource_id, prediction
                )
                for prediction, sample in zip(predictions, small_batch)
            ]
        return return_samples

    def save(self, path):
        self.feature_extractor_new.feature_extractor.save_weights(path + "_features.h5")
        with open(path + "_ilstate.pkl", "wb") as target:
            pkl.dump(self.current_step, target)
        self.classifier.save(path)

    def restore(self, path):
        self.feature_extractor_new.feature_extractor.load_weights(path + "_features.h5")
        with open(path + "_ilstate.pkl", "rb") as target:
            self.current_step = pkl.load(target)
        self.classifier.restore(path)

    def build_image_batch(self, samples, pool=None):
        assert len(samples) > 0
        if pool is not None:
            np_images = pool.map(_get_input_img_np, samples)
        else:
            np_images = [sample.get_resource("input_img_np") for sample in samples]
        return np.stack(np_images, axis=0)


def _get_input_img_np(sample):
    return sample.get_resource("input_img_np")


class KerasBaseModelFactory(components.Factory):
    name_to_class_mapping = KerasBaseModel
    default_section = "keras_base_model"



import multiprocessing
import queue
import threading
import time
from typing import Any, Callable, Generator, Union

from chia import instrumentation


def make_generator_faster(
    gen: Callable[..., Generator[Any, None, None]],
    method: str,
    observable: instrumentation.Observable,
    max_buffer_size: int = 100,
):
    if method == "threading":
        return _make_generator_faster_threading(gen, max_buffer_size, observable)
    elif method == "multiprocessing":
        return _make_generator_faster_multiprocessing(gen, max_buffer_size, observable)
    elif method == "synchronous":
        return gen()
    else:
        raise ValueError(f"Unknown method {method}")


def _producer_main(
    item_queue: Union[queue.Queue, multiprocessing.Queue],
    gen: Callable[..., Generator[Any, None, None]],
    observable: instrumentation.Observable,
) -> None:
    gen_instance = gen()
    for item in gen_instance:
        item_queue.put(item)

    observable.log_debug("Producer done.")
    item_queue.put("THE_END")


def _consumer_generator(
    item_queue: Union[queue.Queue, multiprocessing.Queue],
    observable: instrumentation.Observable,
) -> Generator[Any, None, None]:

    waiting_begin = time.time()
    reported_timeout = False
    while True:
        try:
            item = item_queue.get(timeout=1.0)
            if item == "THE_END":
                observable.log_debug("Consumer done.")
                return
            else:
                yield item
                if reported_timeout:
                    # How long did we wait?
                    waiting_for = time.time() - waiting_begin
                    # Notify user that situation is resolved
                    observable.log_warning(
                        f"Item retrieved with a waiting time of {waiting_for:.2f}s"
                    )

                # Reset waiting timer and reporting
                waiting_begin = time.time()
                reported_timeout = False

        except queue.Empty:
            if not reported_timeout:
                observable.log_warning("Queue timeout hit, waiting for item...")
                reported_timeout = True
            time.sleep(1.0)


def _make_generator_faster_threading(
    gen: Callable[..., Generator[Any, None, None]],
    max_buffer_size: int,
    observable: instrumentation.Observable,
):
    item_queue = queue.Queue(maxsize=max_buffer_size)
    producer_thread = threading.Thread(
        target=_producer_main, args=(item_queue, gen, observable)
    )

    producer_thread.start()
    return _consumer_generator(item_queue, observable)


def _make_generator_faster_multiprocessing(
    gen: Callable[..., Generator[Any, None, None]],
    max_buffer_size: int,
    observable: instrumentation.Observable,
):
    item_queue = multiprocessing.Queue(maxsize=max_buffer_size)
    producer_process = multiprocessing.Process(
        target=_producer_main, args=(item_queue, gen, observable)
    )

    producer_process.start()
    return _consumer_generator(item_queue, observable)



import json
import os.path


def _load_system_config():
    chiarc_filename = os.path.expanduser("~/.chiarc")
    if os.path.exists(chiarc_filename):
        with open(chiarc_filename) as chiarc_file:
            chiarc_content = json.load(chiarc_file)
            assert isinstance(chiarc_content, dict)
            return chiarc_content


__all__ = []



from chia import components
from chia.components.datasets import (
    core50_dataset,
    icifar_dataset,
    icubworld28_dataset,
    ilsvrc2012_dataset,
    inaturalist2018_dataset,
    lndw_dataset,
    nabirds_dataset,
)


class DatasetFactory(components.Factory):
    name_to_class_mapping = {
        "core50": core50_dataset.CORe50Dataset,
        "icifar": icifar_dataset.iCIFARDataset,
        "icubworld28": icubworld28_dataset.iCubWorld28Dataset,
        "inaturalist2018": inaturalist2018_dataset.iNaturalist2018Dataset,
        "ilsvrc2012": ilsvrc2012_dataset.ILSVRC2012Dataset,
        "lndw": lndw_dataset.LNDWDataset,
        "nabirds": nabirds_dataset.NABirdsDataset,
    }
    default_section = "dataset"



from chia.instrumentation.exception_shroud import (
    ExceptionShroud,
    ExceptionShroudFactory,
)
from chia.instrumentation.message import (
    ConfigMessage,
    LogMessage,
    Message,
    MetricMessage,
)
from chia.instrumentation.observable import NamedObservable, Observable
from chia.instrumentation.observers import Observer, ObserverFactory

__all__ = [
    "Message",
    "ConfigMessage",
    "LogMessage",
    "MetricMessage",
    "Observable",
    "NamedObservable",
    "Observer",
    "ObserverFactory",
    "ExceptionShroud",
    "ExceptionShroudFactory",
]



import numpy as np
import tensorflow as tf

from chia import components


class KerasPreprocessor:
    def __init__(
        self,
        augmentation,
        random_crop_to_size=None,
        channel_mean=(127.5, 127.5, 127.5),
        channel_stddev=(127.5, 127.5, 127.5),
    ):
        # Preprocessing
        self.random_crop_to_size = random_crop_to_size
        _channel_mean = channel_mean
        self.channel_mean_normalized = np.array(_channel_mean) / 255.0
        _channel_stddev = channel_stddev
        self.channel_stddev_normalized = np.array(_channel_stddev) / 255.0

        # Augmentation
        self.augmentation = augmentation

    def preprocess_image_batch(self, image_batch, is_training):
        image_batch = tf.cast(image_batch, dtype=tf.float32) / 255.0

        if is_training:
            # Processing_fn expects values in [0, 1]
            image_batch = self.augmentation.process(image_batch)

        # Map to correct range, e.g. [-1.0 , 1.0]
        image_batch = image_batch - self.channel_mean_normalized
        image_batch = image_batch / self.channel_stddev_normalized

        # Do cropping here instead of in augmentation because all augmentation is
        # disabled during testing...
        if self.random_crop_to_size is not None:
            if is_training:
                image_batch = tf.map_fn(self._random_crop_single_image, image_batch)
            else:
                image_batch = tf.image.crop_to_bounding_box(
                    image_batch,
                    (image_batch.shape[1] - self.random_crop_to_size[1]) // 2,
                    (image_batch.shape[2] - self.random_crop_to_size[0]) // 2,
                    self.random_crop_to_size[1],
                    self.random_crop_to_size[0],
                )

        return image_batch

    def _random_crop_single_image(self, image):
        return tf.image.random_crop(
            image, [self.random_crop_to_size[1], self.random_crop_to_size[0], 3]
        )


class KerasPreprocessorFactory(components.Factory):
    name_to_class_mapping = KerasPreprocessor
    default_section = "keras_preprocessor"



import typing
from logging import DEBUG, ERROR, FATAL, INFO, WARNING

from chia.instrumentation.message import (
    LogMessage,
    Message,
    MetricMessage,
    ResultMessage,
    ShutdownMessage,
)
from chia.instrumentation.observers.observer import Observer


class Observable:
    def __init__(self):
        self._observers: typing.List[Observer] = list()

    def _sender_name(self):
        return self.__class__.__name__

    def register(self, observer: Observer):
        self._observers += [observer]

    def unregister(self, observer: Observer):
        self._observers.remove(observer)

    def notify(self, message: Message):
        for observer in self._observers:
            observer.update(message)

    # Helper functions for specific message types
    # Logging
    def log(self, level: int, message: str):
        self.notify(LogMessage(self._sender_name(), level, message))

    def log_debug(self, message: str):
        self.log(DEBUG, message)

    def log_info(self, message: str):
        self.log(INFO, message)

    def log_warning(self, message: str):
        self.log(WARNING, message)

    def log_error(self, message: str):
        self.log(ERROR, message)

    def log_fatal(self, message: str):
        self.log(FATAL, message)

    # Metrics
    def report_metric(self, metric: str, value: float, step: int = -1):
        self.notify(MetricMessage(self._sender_name(), metric, value, step))

    # Results
    def report_result(self, result_dict: dict, step: int = -1):
        self.notify(ResultMessage(self._sender_name(), result_dict, step))

    # Shutdown
    def send_shutdown(self, successful):
        self.notify(ShutdownMessage(self._sender_name(), successful=successful))


class NamedObservable(Observable):
    def __init__(self, custom_sender_name: str):
        super().__init__()
        self._custom_sender_name = custom_sender_name

    def _sender_name(self):
        return self._custom_sender_name



import time

from PIL import Image

from chia import instrumentation


class NetworkResistantImage:
    @staticmethod
    def open(
        filename: str,
        observable: instrumentation.Observable,
        cumulative_wait_max=7200.0,
    ):
        wait_interval_initial = 0.5

        # Set initial state
        wait_interval = wait_interval_initial
        cumulative_wait = 0.0
        last_exception = None

        while cumulative_wait <= cumulative_wait_max:
            try:
                image = Image.open(filename)
                if cumulative_wait > 0.0:
                    observable.log_warning(
                        f"Opened {filename} after waiting for {cumulative_wait} seconds."
                    )
                return image

            except Exception as ex:
                observable.log_warning(
                    f"Cannot open {filename}. Waiting {wait_interval} seconds "
                    f"({type(ex).__name__}: {str(ex)})."
                )
                last_exception = ex
                time.sleep(wait_interval)
                cumulative_wait += wait_interval
                wait_interval *= 2.0

        # After waiting for too long...
        formatted_exception = f"{type(last_exception).__name__}: {str(last_exception)}"
        observable.log_fatal(
            f"Could not open {filename} after waiting for {cumulative_wait} seconds. "
            f"{formatted_exception}"
        )
        raise last_exception



import os

import config as pcfg


def get_user_config():
    chiarc_filename = os.path.expanduser("~/.chiarc")
    if os.path.exists(chiarc_filename):
        return pcfg.config_from_json(chiarc_filename, read_from_file=True)
    else:
        return pcfg.config_from_dict(dict())



from chia import instrumentation
from chia.knowledge.concept import Concept, ConceptFlag
from chia.knowledge.knowledge_base import KnowledgeBase
from chia.knowledge.messages import ConceptChangeMessage, RelationChangeMessage
from chia.knowledge.relation import Relation, RelationFlag


class Discovery(instrumentation.Observer):
    def __init__(self, knowledge_base: KnowledgeBase):
        self.knowledge_base = knowledge_base

    def update(self, message: instrumentation.Message):
        if isinstance(message, ConceptChangeMessage) or isinstance(
            message, RelationChangeMessage
        ):
            for relation in self.knowledge_base.relations():
                self.update_relation(relation)

    def update_relation(self, relation: Relation):
        if self._is_compatible(relation):

            discovered_pairs = set()
            discovered_concepts = set()
            known_concepts = set()
            for concept in self.knowledge_base.concepts():
                discovered_concepts |= {concept.uid}
                known_concepts |= {concept.uid}

                # Explore to the right
                uid_left = concept.uid

                for relation_source in relation.sources:
                    uids_to_right = relation_source.get_right_for(uid_left)
                    discovered_pairs |= {
                        (uid_left, uid_right) for uid_right in uids_to_right
                    }
                    discovered_concepts |= set(uids_to_right)

            # Add relation pairs first. This will trigger another discovery.
            relation.add_pairs(discovered_pairs)

            # Then, add new concepts. This will happen in the second run. The outer run does
            # it again, but it will not have an effect.
            discovered_concepts -= known_concepts
            concepts_for_kb = [
                Concept(uid, {ConceptFlag.AUTO_DISCOVERED})
                for uid in discovered_concepts
            ]
            self.knowledge_base.add_concepts(concepts_for_kb)

        else:
            # Skip relation
            pass

    @staticmethod
    def _is_compatible(relation: Relation) -> bool:
        return (
            RelationFlag.SYMMETRIC not in relation.flags
            and RelationFlag.REFLEXIVE not in relation.flags
            and RelationFlag.TRANSITIVE in relation.flags
            and RelationFlag.EXPLORE_RIGHT in relation.flags
        )



from chia import components
from chia.components.base_models import keras


class BaseModelContainerFactory(components.ContainerFactory):
    name_to_class_mapping = {"keras": keras.KerasBaseModelContainer}
    default_section = "base_model"



from chia import instrumentation


class ConceptChangeMessage(instrumentation.Message):
    def __init__(self, sender: str):
        super().__init__(sender=sender)


class RelationChangeMessage(instrumentation.Message):
    def __init__(self, sender: str):
        super().__init__(sender=sender)



import os


def setup_environment():
    enable_segfault_catching()
    validate_gpu_configuration()
    setup_tensorflow()


def enable_segfault_catching():
    import faulthandler

    faulthandler.enable()
    return True


def validate_gpu_configuration():
    try:
        import os

        from GPUtil import GPUtil

        # Disable GPUS if desired
        if "CHIA_CPU_ONLY" in os.environ.keys():
            if os.environ["CHIA_CPU_ONLY"] == "1":
                print(
                    "Requested CPU only operation."
                    "Disabling all GPUS via environment variable."
                )
                os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
                return True

        gpus = GPUtil.getGPUs()
        if "CUDA_VISIBLE_DEVICES" in os.environ.keys():
            cuda_filter = os.environ["CUDA_VISIBLE_DEVICES"].split(",")
            print(f"Found CUDA device filter: {cuda_filter}")
            gpus = [gpu for gpu in gpus if str(gpu.id) in cuda_filter]

        available_gpus = []
        for gpu in gpus:
            print(f"Found GPU: {gpu.name}")
            if gpu.memoryUtil < 0.5:
                available_gpus += gpus
            else:
                print(
                    "Cannot use this GPU because of its "
                    f"memory utilization @ {int(100.0 * gpu.memoryUtil)}%."
                )

        if len(available_gpus) > 1:
            print(
                f"Only one GPU is supported right now, found {len(available_gpus)} available."
            )
            return False

        if len(available_gpus) < 1:
            print("Need an available GPU!")
            return False

    except Exception:
        print("Could not read available VRAM.")

    return True


def setup_tensorflow():
    # Try turning off tensorflow warnings
    # Sadly, this is broken again in 2.3 :(
    # https://github.com/tensorflow/tensorflow/issues/31870
    # To fix it, you need to call this method before importing anything
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"

    # Fix bug with tf 2.1.0 where cuDNN fails to initialize
    # See https://github.com/tensorflow/tensorflow/issues/24496
    os.environ["TF_FORCE_GPU_ALLOW_GROWTH"] = "true"

    return True


class EnvironmentInfo:
    def __init__(self):
        import chia

        self.chia_version = chia.__version__
        self.environment_variables = dict(os.environ)



import time

from chia.components.runners.runner import Runner


class EpochRunner(Runner):
    def __init__(self, experiment_container, epochs, max_test_samples=None):
        super().__init__(experiment_container)
        self.epochs = epochs
        self.max_test_samples = max_test_samples

    def run(self):
        # Get out some members of container
        dataset = self.experiment_container.dataset
        base_model = self.experiment_container.base_model

        # Build training data
        self.log_info("Loading training pool 0...")
        training_samples = dataset.train_pool(0, "label_gt")

        # Build test data
        self.log_info("Loading testing pool 0...")
        if self.max_test_samples is not None:
            test_samples = dataset.test_pool(0, "label_gt")[: self.max_test_samples]
        else:
            test_samples = dataset.test_pool(0, "label_gt")

        # "Interact"
        self.log_info("Performing interaction...")
        training_samples = self.experiment_container.interactor.query_annotations_for(
            training_samples, "label_gt", "label_ann"
        )

        for epoch in range(self.epochs):
            epoch_begin_time = time.time()

            self.log_info(f"Start of epoch {epoch + 1} of {self.epochs}")
            self.log_info("Observing training data...")
            base_model.observe(training_samples, "label_ann")

            self.log_info("Predicting test data...")
            predicted_test_samples = base_model.predict(test_samples, "label_pred")

            # Go over all evaluators
            self.log_info("Evaluating predicted test data...")
            result_dict = {}
            for evaluator in self.experiment_container.evaluators:
                evaluator.update(predicted_test_samples, "label_gt", "label_pred")
                result_dict.update(evaluator.result())
                evaluator.reset()

            # Store epoch duration
            epoch_end_time = time.time()
            epoch_duration = epoch_end_time - epoch_begin_time
            result_dict.update({"epoch_duration": epoch_duration})

            self.report_result(result_dict, step=epoch + 1)
            self.log_info("Epoch done.")



import os

import imageio

from chia import data
from chia.components.datasets import dataset
from chia.knowledge import relation

_namespace_uid = "iCubWorld28"

_icubworld28_labels_to_wordnet = [
    ("cup", "cup.n.01"),
    ("dishwashing-detergent", "dishwasher_detergent.n.01"),
    ("laundry-detergent", "laundry_detergent.n.01"),
    ("plate", "plate.n.04"),
    ("soap", "bar_soap.n.01"),
    ("sponge", "sponge.n.01"),
    ("sprayer", "atomizer.n.01"),
]


class iCubWorld28Dataset(dataset.Dataset):
    def __init__(self, base_path):
        self.base_path = base_path

    def setup(self, **kwargs):
        pass

    def train_pool_count(self):
        return 4

    def test_pool_count(self):
        return 4

    def train_pool(self, index, label_resource_id):
        return self.get_train_pool_for(index + 1, label_resource_id)

    def test_pool(self, index, label_resource_id):
        return self.get_test_pool_for(index + 1, label_resource_id)

    def namespace(self):
        return _namespace_uid

    def get_train_pool_for(self, day, label_resource_id):
        return self._build_samples("train", day, label_resource_id)

    def get_test_pool_for(self, day, label_resource_id):
        return self._build_samples("test", day, label_resource_id)

    def _build_samples(self, split, day, label_resource_id):
        assert day > 0
        samples = []
        for (category, wordnet_synset) in _icubworld28_labels_to_wordnet:
            for individual in range(1, 4 + 1):
                sample_dir = os.path.join(
                    self.base_path,
                    split,
                    f"day{day}",
                    category,
                    f"{category}{individual}",
                )
                for filename in sorted(os.listdir(sample_dir)):
                    samples += [
                        data.Sample(
                            source=self.__class__.__name__,
                            uid=f"{_namespace_uid}::{split}:{day}:"
                            + f"{category}{individual}:{filename}",
                        )
                        .add_resource(
                            self.__class__.__name__,
                            label_resource_id,
                            f"{_namespace_uid}::{category}{individual}",
                        )
                        .add_resource(
                            self.__class__.__name__,
                            "image_input_np",
                            imageio.imread(os.path.join(sample_dir, filename)),
                        )
                    ]
        return samples

    def get_hyponymy_relation_source(self):
        relation_ = []
        for label, synset in _icubworld28_labels_to_wordnet:
            for individual in range(1, 4 + 1):
                relation_ += [
                    (f"{_namespace_uid}::{label}{individual}", f"WordNet3.0::{synset}")
                ]
        return relation.StaticRelationSource(relation_)

    def prediction_targets(self):
        return [
            f"{_namespace_uid}::{concept_uid}{individual}"
            for individual in range(1, 4 + 1)
            for concept_uid, _ in _icubworld28_labels_to_wordnet
        ]



from chia.knowledge.concept import Concept, ConceptFlag
from chia.knowledge.knowledge_base import KnowledgeBase, KnowledgeBaseFactory
from chia.knowledge.messages import ConceptChangeMessage, RelationChangeMessage
from chia.knowledge.relation import Relation, RelationFlag, RelationSource
from chia.knowledge.wordnet import WordNetAccess

__all__ = [
    "KnowledgeBase",
    "KnowledgeBaseFactory",
    "Concept",
    "ConceptFlag",
    "ConceptChangeMessage",
    "Relation",
    "RelationFlag",
    "RelationSource",
    "RelationChangeMessage",
    "WordNetAccess",
]



import enum
import typing


class ConceptFlag(enum.Enum):
    PREDICTION_TARGET = enum.auto()
    AUTO_DISCOVERED = enum.auto()
    OBSERVED = enum.auto()


class Concept:
    def __init__(self, uid: str, flags: typing.Set[ConceptFlag] = None):
        self.uid = uid
        if flags is not None:
            self.flags = flags
        else:
            self.flags = set()



import csv
import os

import numpy as np
from PIL import Image

from chia import data
from chia.components.datasets import dataset
from chia.knowledge import relation

_namespace_uid = "LNDW"


class LNDWDataset(dataset.Dataset):
    def __init__(self, base_path, side_length, viability_threshold=3.0):
        self.base_path = base_path
        self.side_length = side_length
        self._viability_threshold = viability_threshold

        self.all_classes_even_unviable = []
        with open(os.path.join(self.base_path, "classes.csv")) as classes_file:
            reader = csv.reader(classes_file, delimiter=";")
            header = next(reader)
            fields = {
                "folder": "ID",
                "class_name": "Class Name",
                "individual_id": "No.",
                "grade": "Grade",
                "wordnet": "WordNet",
            }
            fields = {k: header.index(v) for k, v in fields.items()}
            for line in reader:
                self.all_classes_even_unviable += [
                    {k: line[v] for k, v in fields.items()}
                ]

            self.viable_classes = [
                class_
                for class_ in self.all_classes_even_unviable
                if self._viable(class_)
            ]

        self.wordnet_mapping = []
        for class_ in self.viable_classes:
            self.wordnet_mapping += [
                (
                    f"{_namespace_uid}::{class_['class_name']}"
                    + f"{int(class_['individual_id']):02d}",
                    f"WordNet3.0::{class_['wordnet']}",
                )
            ]
            self.wordnet_mapping += [
                (
                    f"{_namespace_uid}::{class_['class_name']}",
                    f"WordNet3.0::{class_['wordnet']}",
                )
            ]

        # Attributes are set in setup()
        self.individuals = None
        self.setup()

    def setup(self, individuals=False, **kwargs):
        self.individuals = individuals

    def setups(self):
        return [{"individuals": True}, {"individuals": False}]

    def train_pool_count(self):
        return 1

    def test_pool_count(self):
        return 1

    def train_pool(self, index, label_resource_id):
        assert index == 0
        return self.get_train_pool(label_resource_id, self.individuals)

    def test_pool(self, index, label_resource_id):
        assert index == 0
        return self.get_test_pool(label_resource_id, self.individuals)

    def namespace(self):
        return _namespace_uid

    def _viable(self, class_):
        return float(class_["grade"]) <= self._viability_threshold

    def _filenames(self):
        return [
            "01W.jpg",
            "02SW.jpg",
            "03S.jpg",
            "04SE.jpg",
            "05E.jpg",
            "06NE.jpg",
            "07N.jpg",
            "08NW.jpg",
            "09TOP.jpg",
        ]

    def _build_sample(self, class_, filename, label_resource_id, individuals):
        # Open and resize image
        the_image = Image.open(
            os.path.join(self.base_path, f"{int(class_['folder']):02d}", filename)
        )
        the_image = np.asarray(
            the_image.resize((self.side_length, self.side_length), Image.ANTIALIAS)
        )

        if individuals:
            label_string = (
                f"{_namespace_uid}::{class_['class_name']}"
                + f"{int(class_['individual_id']):02d}"
            )
        else:
            label_string = f"{_namespace_uid}::{class_['class_name']}"

        # Build sample
        the_sample = (
            data.Sample(
                source=self.__class__.__name__,
                uid=f"{_namespace_uid}::{class_}.{filename}",
            )
            .add_resource(self.__class__.__name__, "input_img_np", the_image)
            .add_resource(self.__class__.__name__, label_resource_id, label_string)
        )
        return the_sample

    def get_train_pool(self, label_resource_id, individuals=False):
        samples = []
        for class_ in self.viable_classes:
            for filename in self._filenames()[:8]:
                the_sample = self._build_sample(
                    class_, filename, label_resource_id, individuals
                )
                samples += [the_sample]

        return samples

    def get_test_pool(self, label_resource_id, individuals=False):
        samples = []
        for class_ in self.viable_classes:
            for filename in self._filenames()[8:]:
                the_sample = self._build_sample(
                    class_, filename, label_resource_id, individuals
                )
                samples += [the_sample]

        return samples

    def get_hyponymy_relation_source(self):
        return relation.StaticRelationSource(self.wordnet_mapping)

    def prediction_targets(self):
        if self.individuals:
            return {
                f"{_namespace_uid}::{class_['class_name']}"
                + f"{int(class_['individual_id']):02d}"
                for class_ in self.viable_classes
            }
        else:
            return {
                f"{_namespace_uid}::{class_['class_name']}"
                for class_ in self.viable_classes
            }



import logging
import time
import typing


class Message:
    def __init__(self, sender: str):
        self.sender = sender
        self.timestamp: float = time.time()

    def _format_timestamp(self) -> str:
        return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.timestamp))

    def __str__(self):
        return f"[{self._format_timestamp()}] [MESSAGE] [{self.sender}]: {self.__class__.__name__}"


class LogMessage(Message):
    def __init__(self, sender: str, level: int, message: str):
        super().__init__(sender=sender)
        self.sender = sender
        self.level = level
        self.message = message

    def __str__(self):
        return (
            f"[{self._format_timestamp()}] [{logging.getLevelName(self.level)}] "
            f"[{self.sender}]: {self.message}"
        )


class ConfigMessage(Message):
    def __init__(self, sender: str, field: str, value: typing.Any, source: str):
        super().__init__(sender=sender)
        self.field = field
        self.value = value
        self.source = source

    def __str__(self):
        return (
            f"[{self._format_timestamp()}] [CONFIGURATION] [{self.sender}]: "
            f"config field {self.field} set to {self.value} from {self.source}"
        )


class MetricMessage(Message):
    def __init__(self, sender: str, metric: str, value: float, step: int):
        super().__init__(sender=sender)
        self.metric = metric
        self.value = value
        self.step = step

    def __str__(self):
        return (
            f"[{self._format_timestamp()}] [METRIC] [{self.sender}]: "
            f"metric {self.metric} @{self.step} = {self.value}"
        )


class ResultMessage(Message):
    def __init__(self, sender: str, result_dict: dict, step: int):
        super().__init__(sender=sender)
        self.result_dict = result_dict
        self.step = step

    def __str__(self):
        return (
            f"[{self._format_timestamp()}] [RESULT] [{self.sender}]: "
            f"keys are {','.join(self.result_dict.keys())} @{self.step}"
        )


class ShutdownMessage(Message):
    """This message tells the observers that they should save their data."""

    def __init__(self, sender: str, successful: bool):
        super().__init__(sender)
        self.successful = successful

    def __str__(self):
        return (
            f"[{self._format_timestamp()}] [SHUTDOWN] [{self.sender}] "
            f"Successful: {self.successful}"
        )



import random
import typing

from chia import helpers, instrumentation, knowledge
from chia.components import datasets, evaluators, interactors, runners
from chia.components.base_models import incremental_model
from chia.components.datasets.dataset import Dataset
from chia.components.interactors import interactor
from chia.containers import model


class ExperimentMetadata(typing.NamedTuple):
    name: str
    run_id: str


class ExperimentContainer:
    def __init__(self, config, outer_observable: instrumentation.Observable):
        # Configuration
        self.config = config

        # Experiment metadata
        self.experiment_metadata = ExperimentMetadata(
            name=config["meta"]["name"], run_id=("%08x" % random.randrange(2 ** 32))
        )

        # Environment info
        self.environment_info = helpers.EnvironmentInfo()

        # Load all observers
        self.observers: typing.List[instrumentation.Observer] = [
            instrumentation.ObserverFactory.create(
                sub_config,
                experiment_metadata=self.experiment_metadata,
                environment_info=self.environment_info,
            )
            for sub_config in config["observers"]
        ]
        for observer in self.observers:
            outer_observable.register(observer)

        # Make an exception shroud
        self.exception_shroud = instrumentation.ExceptionShroudFactory.create(
            dict(), observers=self.observers
        )

        with self.exception_shroud:
            # Build knowledge base
            self.knowledge_base = knowledge.KnowledgeBaseFactory.create(
                dict(), observers=self.observers
            )

            # Dataset
            self.dataset: Dataset = datasets.DatasetFactory.create(
                config["dataset"], observers=self.observers
            )

            # Interactor
            self.interactor: interactor.Interactor = interactors.InteractorFactory.create(
                config["interactor"], kb=self.knowledge_base, observers=self.observers
            )

            # Model container
            self.model_container = model.ModelContainer(
                config["model"],
                knowledge_base=self.knowledge_base,
                observers=self.observers,
            )

            # Instantiate all evaluators
            self.evaluators: typing.List[evaluators.Evaluator] = [
                evaluators.EvaluatorFactory.create(
                    sub_config, kb=self.knowledge_base, observers=self.observers
                )
                for sub_config in config["evaluators"]
            ]

            # Allow outside access
            self.classifier = self.model_container.classifier
            self.base_model_container = self.model_container.base_model_container
            self.base_model: incremental_model.IncrementalModel = self.model_container.base_model_container.base_model

            # Run KB setup
            self._setup_knowledge_base()

            # Build runner
            self.runner: runners.Runner = runners.RunnerFactory.create(
                config["runner"], experiment_container=self, observers=self.observers
            )

    def _setup_knowledge_base(self):
        """Sets up the prediction targets and hyponymy relation from the dataset."""
        # Get prediction targets
        self.knowledge_base.add_prediction_targets(self.dataset.prediction_targets())

        # Add relation sources
        relation_sources = [self.dataset.get_hyponymy_relation_source()]
        if self.config["with_wordnet"]:
            relation_sources += [knowledge.WordNetAccess()]
        self.knowledge_base.add_hyponymy_relation(relation_sources)



"""paths: Find files in default paths"""
import os

from chia import instrumentation


def maybe_expand_path(path, observable: instrumentation.Observable):
    """Replaces the argument path with a default one if necessary and possible.

    For example, if the user specifies inat_features.h5 and it is not in the CWD,
    we look it up in ~/.chia/ and return this path if the file exists there."""
    observable.log_debug(f"Expanding path for: {path}")

    # (1) Check if path is valid already
    if os.path.exists(path):
        observable.log_debug("  found in CWD.")
        return path

    # (2) Check if path is absolute (even if invalid)
    if os.path.isabs(path):
        observable.log_debug("  is absolute and will not be processed further.")
        return path

    # (3) Check if path is valid in ~/.chia
    chia_directory = os.path.expanduser("~/.chia")
    joined_path = os.path.join(chia_directory, path)
    if os.path.exists(joined_path):
        observable.log_debug(f"  found as {joined_path} in user directory")
        return joined_path

    observable.log_debug("  not found, not processing.")
    return path



import abc
import enum
import typing

import networkx as nx

from chia import instrumentation
from chia.knowledge.messages import RelationChangeMessage


class RelationSource(abc.ABC):
    @abc.abstractmethod
    def get_left_for(self, uid_right):
        pass

    @abc.abstractmethod
    def get_right_for(self, uid_left):
        pass


class StaticRelationSource:
    def __init__(self, data):
        self.right_for_left = {}
        self.left_for_right = {}
        for (left, right) in data:
            if left in self.right_for_left.keys():
                self.right_for_left[left] += [right]
            else:
                self.right_for_left[left] = [right]

            if right in self.left_for_right.keys():
                self.left_for_right[right] += [left]
            else:
                self.left_for_right[right] = [left]

    def get_left_for(self, uid_right):
        if uid_right in self.left_for_right.keys():
            return self.left_for_right[uid_right]
        else:
            return set()

    def get_right_for(self, uid_left):
        if uid_left in self.right_for_left.keys():
            return self.right_for_left[uid_left]
        else:
            return set()


class RelationFlag(enum.Enum):
    HYPONYMY = enum.auto()
    SYMMETRIC = enum.auto()
    REFLEXIVE = enum.auto()
    TRANSITIVE = enum.auto()
    EXPLORE_LEFT = enum.auto()
    EXPLORE_RIGHT = enum.auto()


class Relation(instrumentation.Observable):
    def __init__(
        self,
        uid: str,
        sources: typing.List[RelationSource],
        flags: typing.Optional[typing.Set[RelationFlag]] = None,
    ):
        instrumentation.Observable.__init__(self)
        self.uid: str = uid
        self.sources: typing.List[RelationSource] = sources

        if flags is not None:
            self.flags: typing.Set[RelationFlag] = flags
        else:
            self.flags: typing.Set[RelationFlag] = set()

        self._rgraph: nx.DiGraph = nx.DiGraph()
        self._ugraph: nx.Graph = nx.Graph()
        self._pairs: typing.Set[typing.Tuple[str, str]] = set()

    def pairs(self) -> typing.Set[typing.Tuple[str, str]]:
        return self._pairs

    def add_pairs(self, pairs: typing.Set[typing.Tuple[str, str]]):
        changed = False
        for pair in pairs:
            changed |= self._add_pair(pair)

        if changed:
            message = RelationChangeMessage(self._sender_name())
            self.notify(message)

    def _add_pair(self, pair: typing.Tuple[str, str]):
        if pair not in self._pairs:
            self._pairs |= {pair}

            # This is reversed on purpose
            self._rgraph.add_edge(pair[1], pair[0])
            self._ugraph.add_edge(pair[1], pair[0])
            return True
        else:
            return False

    def rgraph(self) -> nx.DiGraph:
        return self._rgraph

    def ugraph(self) -> nx.Graph:
        return self._ugraph



from typing import Optional

import networkx as nx
import numpy as np

from chia import instrumentation, knowledge
from chia.components.interactors import interactor


class NoisyOracleInteractor(
    interactor.Interactor, instrumentation.Observable, instrumentation.Observer
):
    def __init__(
        self,
        kb,
        noise_model,
        inaccuracy=0.0,
        relabel_fraction=None,
        lambda_=None,
        q=None,
        filter_imprecise=False,
        project_to_random_leaf=False,
    ):
        interactor.Interactor.__init__(self, kb=kb)
        instrumentation.Observable.__init__(self)
        instrumentation.Observer.__init__(self)

        self.noise_model = noise_model
        self.inaccuracy = inaccuracy

        if self.noise_model == "Deng2014":
            assert relabel_fraction is not None
            self.relabel_fraction: float = relabel_fraction
        elif self.noise_model == "Poisson":
            assert lambda_ is not None
            self.lambda_: float = lambda_
        elif self.noise_model == "Geometric":
            assert q is not None
            self.q: float = q
        elif self.noise_model == "Inaccuracy":
            pass
        else:
            raise ValueError(f"Unknown noise model: {self.noise_model}")

        self.filter_imprecise = filter_imprecise
        self.project_to_random_leaf = project_to_random_leaf

        self.is_updated = False
        self.graph: Optional[nx.DiGraph] = None
        self.root = None
        self.leaf_nodes = None

        self._kb.register(self)

    def _apply_deng_noise(self, uid):
        if np.random.binomial(1, self.relabel_fraction):
            chosen_predecessor = next(
                self.graph.predecessors(uid)
            )  # TODO what to do if there is more than 1 parent?
            return chosen_predecessor
        else:
            return uid

    def _apply_geometric_noise(self, uid):
        target = np.random.geometric(1 - self.q) - 1
        return self._reduce_depth_to(uid, target)

    def _apply_poisson_noise(self, uid):
        target = np.random.poisson(self.lambda_)
        return self._reduce_depth_to(uid, target)

    def _reduce_depth_to(self, uid, depth_target):
        path_to_label = nx.shortest_path(self.graph, self.root, uid)
        final_depth = max(0, min(len(path_to_label) - 1, depth_target))
        return path_to_label[final_depth]

    def _project_to_random_leaf(self, uid):
        if self.graph.out_degree(uid) == 0:  # noqa
            return uid
        else:
            # List all descendants
            all_descendants = nx.descendants(self.graph, uid)

            # Use only leaves
            valid_descendants = list(
                filter(lambda n: self.graph.out_degree(n) == 0, all_descendants)  # noqa
            )

            return np.random.choice(valid_descendants)

    def _maybe_update_graphs(self):
        if not self.is_updated:
            try:
                self.graph = self._kb.get_hyponymy_relation_rgraph()
                self.root = next(nx.topological_sort(self.graph))
                self.leaf_nodes = list(
                    filter(
                        lambda n: self.graph.out_degree(n) == 0, self.graph.nodes
                    )  # noqa
                )
                self.is_updated = True
            except ValueError:
                # No graph available yet
                pass

    def query_annotations_for(self, samples, gt_resource_id, ann_resource_id):
        self._maybe_update_graphs()

        # Add noise
        noisy_samples = [
            sample.add_resource(
                self.__class__.__name__,
                ann_resource_id,
                self.apply_noise(sample.get_resource(gt_resource_id)),
            )
            for sample in samples
        ]

        # Count modified samples
        modified_samples = sum(
            [
                1
                if noisy_sample.get_resource(gt_resource_id)
                != noisy_sample.get_resource(ann_resource_id)
                else 0
                for noisy_sample in noisy_samples
            ]
        )
        self.log_debug(f"Modified {modified_samples} out of {len(samples)} samples.")

        # Filter imprecise samples
        precise_only_samples = [
            sample
            for sample in noisy_samples
            if self.apply_filter(sample.get_resource(ann_resource_id))
        ]
        self.log_debug(
            f"Filtered out {len(samples)-len(precise_only_samples)}"
            + f" out of {len(samples)} samples."
        )
        return precise_only_samples

    def apply_noise(self, uid):
        # Apply inaccuracy
        if np.random.uniform() <= self.inaccuracy:
            assert uid in self.leaf_nodes
            inaccurate_uid = np.random.choice(self.leaf_nodes)
        else:
            inaccurate_uid = uid

        # Select noise model
        if self.noise_model == "Deng2014":
            noisy_uid = self._apply_deng_noise(inaccurate_uid)
        elif self.noise_model == "Geometric":
            noisy_uid = self._apply_geometric_noise(inaccurate_uid)
        elif self.noise_model == "Poisson":
            noisy_uid = self._apply_poisson_noise(inaccurate_uid)
        elif self.noise_model == "Inaccuracy":
            noisy_uid = inaccurate_uid
        else:
            raise ValueError(f"Unknown noise model {self.noise_model}")

        # Project to random leaf
        if self.project_to_random_leaf:
            noisy_uid = self._project_to_random_leaf(noisy_uid)

        return noisy_uid

    def apply_filter(self, uid):
        if self.filter_imprecise:
            return self.graph.out_degree(uid) == 0  # noqa
        else:
            return True

    def update(self, message: instrumentation.Message):
        if isinstance(message, knowledge.RelationChangeMessage) or isinstance(
            message, knowledge.ConceptChangeMessage
        ):
            self.is_updated = False



from chia.helpers.batches import batches_from, batches_from_pair
from chia.helpers.environment import EnvironmentInfo, setup_environment
from chia.helpers.ioqueue import make_generator_faster
from chia.helpers.paths import maybe_expand_path
from chia.helpers.robustness import NetworkResistantImage
from chia.helpers.user_config import get_user_config

__all__ = [
    "batches_from",
    "batches_from_pair",
    "make_generator_faster",
    "maybe_expand_path",
    "NetworkResistantImage",
    "setup_environment",
    "EnvironmentInfo",
    "get_user_config",
]



import abc

from chia import knowledge


class Dataset(abc.ABC):
    @abc.abstractmethod
    def setup(self, **kwargs):
        pass

    def setups(self):
        return [dict()]

    @abc.abstractmethod
    def train_pool_count(self):
        pass

    @abc.abstractmethod
    def test_pool_count(self):
        pass

    @abc.abstractmethod
    def train_pool(self, index, label_resource_id):
        pass

    @abc.abstractmethod
    def test_pool(self, index, label_resource_id):
        pass

    @abc.abstractmethod
    def namespace(self):
        pass

    @abc.abstractmethod
    def get_hyponymy_relation_source(self) -> knowledge.RelationSource:
        pass

    @abc.abstractmethod
    def prediction_targets(self):
        pass



import traceback

from chia import components
from chia.instrumentation import observable


class ExceptionShroud(observable.Observable):
    def __init__(self):
        observable.Observable.__init__(self)

    def __enter__(self):
        self.log_debug("Entering exception shroud")

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.log_debug("Leaving exception shroud with exception")

            self.log_fatal(
                "".join(traceback.format_exception(exc_type, exc_val, exc_tb))
            )
            self.send_shutdown(successful=False)
        else:
            self.log_debug("Leaving exception shroud without exception")

        return False


class ExceptionShroudFactory(components.Factory):
    name_to_class_mapping = ExceptionShroud



from chia import components
from chia.components.classifiers import keras_idk_hc, keras_onehot_hc


class ClassifierFactory(components.Factory):
    name_to_class_mapping = {
        "keras_onehot": keras_onehot_hc.OneHotEmbeddingBasedKerasHC,
        "keras_idk": keras_idk_hc.IDKEmbeddingBasedKerasHC,
    }
    default_section = "classifier"



from abc import ABC, abstractmethod

from chia import knowledge


class Interactor(ABC):
    def __init__(self, kb: knowledge.KnowledgeBase):
        self._kb = kb

    @abstractmethod
    def query_annotations_for(self, samples, gt_resource_id, ann_resource_id):
        return None



import glob
import os
import pickle as pkl

import numpy as np

from chia import data
from chia.components.datasets import dataset
from chia.knowledge import relation

_namespace_uid = "CORe50"

_labels_to_wordnet = [
    ("plug_adapter1", "plug.n.05"),
    ("plug_adapter2", "plug.n.05"),
    ("plug_adapter3", "plug.n.05"),
    ("plug_adapter4", "plug.n.05"),
    ("plug_adapter5", "plug.n.05"),
    ("mobile_phone1", "cellular_telephone.n.01"),
    ("mobile_phone2", "cellular_telephone.n.01"),
    ("mobile_phone3", "cellular_telephone.n.01"),
    ("mobile_phone4", "cellular_telephone.n.01"),
    ("mobile_phone5", "cellular_telephone.n.01"),
    ("scissor1", "scissors.n.01"),
    ("scissor2", "scissors.n.01"),
    ("scissor3", "scissors.n.01"),
    ("scissor4", "scissors.n.01"),
    ("scissor5", "scissors.n.01"),
    ("light_bulb1", "light_bulb.n.01"),
    ("light_bulb2", "light_bulb.n.01"),
    ("light_bulb3", "light_bulb.n.01"),
    ("light_bulb4", "light_bulb.n.01"),
    ("light_bulb5", "light_bulb.n.01"),
    ("can1", "soda_can.n.01"),
    ("can2", "soda_can.n.01"),
    ("can3", "soda_can.n.01"),
    ("can4", "soda_can.n.01"),
    ("can5", "soda_can.n.01"),
    ("glass1", "spectacles.n.01"),
    ("glass2", "sunglasses.n.01"),
    ("glass3", "sunglasses.n.01"),
    ("glass4", "sunglasses.n.01"),
    ("glass5", "sunglasses.n.01"),
    ("ball1", "ball.n.01"),
    ("ball2", "tennis_ball.n.01"),
    ("ball3", "football.n.02"),
    ("ball4", "ball.n.01"),
    ("ball5", "football.n.02"),
    ("marker1", "highlighter.n.02"),
    ("marker2", "highlighter.n.02"),
    ("marker3", "highlighter.n.02"),
    ("marker4", "highlighter.n.02"),
    ("marker5", "highlighter.n.02"),
    ("cup1", "cup.n.01"),
    ("cup2", "cup.n.01"),
    ("cup3", "cup.n.01"),
    ("cup4", "cup.n.01"),
    ("cup5", "cup.n.01"),
    ("remote_control1", "remote_control.n.01"),
    ("remote_control2", "remote_control.n.01"),
    ("remote_control3", "remote_control.n.01"),
    ("remote_control4", "remote_control.n.01"),
    ("remote_control5", "remote_control.n.01"),
]


class CORe50Dataset(dataset.Dataset):
    def __init__(self, base_path):
        self.base_path = base_path

        self.imgs = np.load(
            os.path.join(self.base_path, "core50_imgs.npy"), mmap_mode="r"
        )

        with open(os.path.join(self.base_path, "paths.pkl"), "rb") as paths_file:
            self.paths = pkl.load(paths_file)

        with open(
            os.path.join(self.base_path, "labels2names.pkl"), "rb"
        ) as labels_to_names_file:
            self.labels_to_names = pkl.load(labels_to_names_file)

        self.path_to_index = {path: index for index, path in enumerate(self.paths)}

        # Attributes are assigned in setup()
        self.scenario = None
        self.run = None
        self.setup()

    def setups(self):
        setups = []
        for scenario in ["ni", "nc", "nic"]:
            for run in range(self.get_run_count(scenario)):
                setups += [{"scenario": scenario, "run": run}]
        return setups

    def setup(self, scenario="nic", run=0, **kwargs):
        self.scenario = scenario
        self.run = run

    def train_pool_count(self):
        return self.get_train_pool_count(self.scenario, self.run)

    def test_pool_count(self):
        return 1

    def train_pool(self, index, label_resource_id):
        return self.get_pool_for(
            self.scenario, self.run, f"train_batch_{index:02d}", label_resource_id
        )

    def test_pool(self, index, label_resource_id):
        assert index == 0
        return self.get_pool_for(self.scenario, self.run, "test", label_resource_id)

    def namespace(self):
        return _namespace_uid

    def get_train_pool_count(self, scenario, run):
        scenario = str(scenario).lower()
        assert scenario in ["ni", "nc", "nic"]
        filelist_filter = os.path.join(
            self.base_path,
            "batches_filelists",
            f"{scenario.upper()}_inc",
            f"run{run:d}",
            "train_batch_*_filelist.txt",
        )

        return len(glob.glob(filelist_filter))

    def get_run_count(self, scenario):
        scenario = str(scenario).lower()
        assert scenario in ["ni", "nc", "nic"]
        filelist_filter = os.path.join(
            self.base_path, "batches_filelists", f"{scenario.upper()}_inc", "run*"
        )

        return len(glob.glob(filelist_filter))

    def get_pool_for(self, scenario, run, batch, label_resource_id):
        # Find the data
        scenario = str(scenario).lower()
        assert scenario in ["ni", "nc", "nic"]

        filelist_path = os.path.join(
            self.base_path,
            "batches_filelists",
            f"{scenario.upper()}_inc",
            f"run{run:d}",
            f"{batch}_filelist.txt",
        )

        # Find appropriate label map
        label_map = self.labels_to_names[scenario][run]

        samples = []

        with open(filelist_path) as filelist:
            for line in filelist:
                path, class_id = line.strip().split(" ")
                samples += [
                    data.Sample(
                        source=self.__class__.__name__, uid=f"{_namespace_uid}::{path}"
                    )
                    .add_resource(
                        self.__class__.__name__,
                        "input_img_np",
                        self.imgs[self.path_to_index[path]],
                    )
                    .add_resource(
                        self.__class__.__name__,
                        label_resource_id,
                        f"{_namespace_uid}::{label_map[int(class_id)]}",
                    )
                ]

        return samples

    def get_hyponymy_relation_source(self):
        return relation.StaticRelationSource(
            [
                (f"{_namespace_uid}::{label}", f"WordNet3.0::{synset}")
                for label, synset in _labels_to_wordnet
            ]
        )

    def prediction_targets(self):
        return [
            f"{_namespace_uid}::{concept_uid}" for concept_uid, _ in _labels_to_wordnet
        ]



import pickle

import networkx as nx
import numpy as np
import tensorflow as tf

from chia import instrumentation, knowledge
from chia.components.classifiers import keras_hierarchicalclassification


class IDKEmbeddingBasedKerasHC(
    keras_hierarchicalclassification.EmbeddingBasedKerasHC, instrumentation.Observable
):
    def __init__(
        self,
        kb,
        l2=5e-5,
        mlnp=True,
        normalize_scores=True,
        raw_output=False,
        weighting="default",
        gain_compensation="simple",
    ):
        keras_hierarchicalclassification.EmbeddingBasedKerasHC.__init__(self, kb)
        instrumentation.Observable.__init__(self)

        # Configuration
        self._l2_regularization_coefficient = l2

        self._mlnp = mlnp
        self._normalize_scores = normalize_scores

        self._raw_output = raw_output
        if self._raw_output and (self._mlnp or self._normalize_scores):
            raise ValueError(
                "Cannot use raw output and MLNP or normalization at the same time!"
            )

        self._weighting = weighting
        self._gain_compensation = gain_compensation

        self.fc_layer = None
        self.uid_to_dimension = {}
        self.graph = None
        self.prediction_target_uids = None
        self.topo_sorted_uids = None
        self.loss_weights = None
        self.update_embedding()

    def predict_embedded(self, feature_batch):
        return self.fc_layer(feature_batch)

    def embed(self, labels):
        embedding = np.zeros((len(labels), len(self.uid_to_dimension)))
        for i, label in enumerate(labels):
            if label == "chia::UNCERTAIN":
                embedding[i] = 1.0
            else:
                embedding[i, self.uid_to_dimension[label]] = 1.0
                for ancestor in nx.ancestors(self.graph, label):
                    embedding[i, self.uid_to_dimension[ancestor]] = 1.0

        return embedding

    def deembed_dist(self, embedded_labels):
        return [
            self._deembed_single(embedded_label) for embedded_label in embedded_labels
        ]

    def _deembed_single(self, embedded_label):
        conditional_probabilities = {
            uid: embedded_label[i] for uid, i in self.uid_to_dimension.items()
        }

        if self._raw_output:
            # Directly output conditional probabilities
            return list(conditional_probabilities.items())
        else:
            # Stage 1 calculates the unconditional probabilities
            unconditional_probabilities = {}

            for uid in self.topo_sorted_uids:
                unconditional_probability = conditional_probabilities[uid]

                no_parent_probability = 1.0
                has_parents = False
                for parent in self.graph.predecessors(uid):
                    has_parents = True
                    no_parent_probability *= 1.0 - unconditional_probabilities[parent]

                if has_parents:
                    unconditional_probability *= 1.0 - no_parent_probability

                unconditional_probabilities[uid] = unconditional_probability

            # Stage 2 calculates the joint probability of the synset and "no children"
            joint_probabilities = {}
            for uid in reversed(self.topo_sorted_uids):
                joint_probability = unconditional_probabilities[uid]
                no_child_probability = 1.0
                for child in self.graph.successors(uid):
                    no_child_probability *= 1.0 - unconditional_probabilities[child]

                joint_probabilities[uid] = joint_probability * no_child_probability

            tuples = joint_probabilities.items()
            sorted_tuples = list(sorted(tuples, key=lambda tup: tup[1], reverse=True))

            if self._mlnp:
                for i, (uid, p) in enumerate(sorted_tuples):
                    if uid not in self.prediction_target_uids:
                        sorted_tuples[i] = (uid, 0.0)

            if self._normalize_scores:
                total_scores = sum([p for uid, p in sorted_tuples])
                sorted_tuples = [(uid, p / total_scores) for uid, p in sorted_tuples]

            return list(sorted_tuples)

    def update_embedding(self):
        current_concepts = self.kb.concepts()
        current_concept_count = len(current_concepts)
        self.report_metric("current_concepts", current_concept_count)

        if current_concept_count == 0:
            return True

        try:
            old_weights = self.fc_layer.get_weights()
            old_uidtodim = self.uid_to_dimension
            old_graph = self.graph

        except Exception:
            old_weights = []
            old_uidtodim = []
            old_graph = None

        self.fc_layer = tf.keras.layers.Dense(
            current_concept_count,
            activation="sigmoid",
            kernel_regularizer=tf.keras.regularizers.l2(
                self._l2_regularization_coefficient
            )
            if self._l2_regularization_coefficient > 0.0
            else None,
            kernel_initializer="zero",
            bias_initializer="zero",
        )

        try:
            self.graph = self.kb.get_hyponymy_relation_rgraph()
        except ValueError:
            return False

        # Memorize topological sorting for later
        all_uids = nx.topological_sort(self.graph)
        self.topo_sorted_uids = list(all_uids)
        assert len(current_concepts) == len(self.topo_sorted_uids)

        self.uid_to_dimension = {
            uid: dimension for dimension, uid in enumerate(self.topo_sorted_uids)
        }

        self.prediction_target_uids = {
            concept.uid
            for concept in self.kb.concepts(
                flags={knowledge.ConceptFlag.PREDICTION_TARGET}
            )
        }

        if len(old_weights) == 2:
            # Layer can be updated
            new_weights = np.zeros([old_weights[0].shape[0], current_concept_count])
            new_biases = np.zeros([current_concept_count])

            reused_concepts = 0
            for new_uid, dim in self.uid_to_dimension.items():
                # Check if old weight is even available
                if new_uid not in old_uidtodim.keys():
                    continue

                # Check if parents have changed
                if set(self.graph.predecessors(new_uid)) != set(
                    old_graph.predecessors(new_uid)
                ):
                    continue

                new_weights[:, dim] = old_weights[0][:, old_uidtodim[new_uid]]
                new_biases[dim] = old_weights[1][old_uidtodim[new_uid]]
                reused_concepts += 1

            self.report_metric("reused_concepts", reused_concepts)

            self.fc_layer.build([None, old_weights[0].shape[0]])
            self.fc_layer.set_weights([new_weights, new_biases])

        self.update_loss_weights()
        return True

    def update_loss_weights(self):
        if len(self.prediction_target_uids) == 0:
            self.log_debug("Skipping loss weight update, no concepts found.")
            self.loss_weights = []
            return

        self.log_debug(
            f"Updating loss weights. Strategy: {self._weighting}, "
            f"gain compensation: {self._gain_compensation}"
        )

        # (1) Calculate "natural" weights by assuming uniform distribution
        # over observed concepts
        occurences = {uid: 0 for uid in self.topo_sorted_uids}
        for uid in self.prediction_target_uids:
            affected_uids = {uid}
            affected_uids |= nx.ancestors(self.graph, uid)
            for affected_uid in list(affected_uids):
                affected_uids |= set(self.graph.successors(affected_uid))

            for affected_uid in affected_uids:
                occurences[affected_uid] += 1

        occurrence_vector = np.array([occurences[uid] for uid in self.uid_to_dimension])

        # (2) Calculate weight vector
        if self._weighting == "default":
            self.loss_weights = np.ones(len(self.uid_to_dimension))

        elif self._weighting == "equalize":
            try:
                self.loss_weights = (
                    np.ones(len(self.uid_to_dimension)) / occurrence_vector
                )
            except ZeroDivisionError as err:
                self.log_fatal("Division by zero in equalize loss weighting strategy.")
                raise err

        elif self._weighting == "descendants":
            try:
                # Start with an equal weighting
                self.loss_weights = (
                    np.ones(len(self.uid_to_dimension)) / occurrence_vector
                )

                for i, uid in enumerate(self.uid_to_dimension):
                    self.loss_weights[i] *= (
                        len(nx.descendants(self.graph, uid)) + 1.0
                    )  # Add one for the node itself.
            except ZeroDivisionError as err:
                self.log_fatal(
                    "Division by zero in descendants loss weighting strategy."
                )
                raise err

        elif self._weighting == "reachable_leaf_nodes":
            try:
                # Start with an equal weighting
                self.loss_weights = (
                    np.ones(len(self.uid_to_dimension)) / occurrence_vector
                )

                for i, uid in enumerate(self.uid_to_dimension):
                    descendants = set(nx.descendants(self.graph, uid)) | {uid}
                    reachable_leaf_nodes = descendants.intersection(
                        self.prediction_target_uids
                    )
                    self.loss_weights[i] *= len(reachable_leaf_nodes)

                    # Test if any leaf nodes are reachable
                    if len(reachable_leaf_nodes) == 0:
                        raise ValueError(
                            f"In this hierarchy, the node {uid} cannot reach "
                            "any leaf nodes!"
                        )

            except ZeroDivisionError as err:
                self.log_fatal(
                    "Division by zero in reachable_leaf_nodes loss weighting strategy."
                )
                raise err

        else:
            raise ValueError(f'Unknown loss weighting strategy "{self._weighting}"')

        # Normalize so we don't have to adapt the learning rate a lot.
        if self._gain_compensation == "simple":
            gain = np.mean(self.loss_weights)
        elif self._gain_compensation == "per_element":
            gain = np.mean(self.loss_weights * occurrence_vector) / np.mean(
                occurrence_vector
            )
        else:
            raise ValueError(
                f'Unknown gain compensation setting "{self._gain_compensation}"'
            )

        self.report_metric("gain_from_weighting", gain)
        self.loss_weights /= gain

    def loss(self, feature_batch, ground_truth, global_step):
        if not self.is_updated:
            raise RuntimeError(
                "This classifier is not yet ready to compute a loss. "
                "Check if it has been notified of a hyponymy relation."
            )

        loss_mask = np.zeros((len(ground_truth), len(self.uid_to_dimension)))
        for i, label in enumerate(ground_truth):
            # Loss mask
            loss_mask[i, self.uid_to_dimension[label]] = 1.0

            for ancestor in nx.ancestors(self.graph, label):
                loss_mask[i, self.uid_to_dimension[ancestor]] = 1.0
                for successor in self.graph.successors(ancestor):
                    loss_mask[i, self.uid_to_dimension[successor]] = 1.0
                    # This should also cover the node itself, but we do it anyway

            if not self._mlnp:
                # Learn direct successors in order to "stop"
                # prediction at these nodes.
                # If MLNP is active, then this can be ignored.
                # Because we never want to predict
                # inner nodes, we interpret labels at
                # inner nodes as imprecise labels.
                for successor in self.graph.successors(label):
                    loss_mask[i, self.uid_to_dimension[successor]] = 1.0

        embedding = self.embed(ground_truth)
        prediction = self.predict_embedded(feature_batch)

        # Binary cross entropy loss function
        clipped_probs = tf.clip_by_value(prediction, 1e-7, (1.0 - 1e-7))
        the_loss = -(
            embedding * tf.math.log(clipped_probs)
            + (1.0 - embedding) * tf.math.log(1.0 - clipped_probs)
        )

        sum_per_batch_element = tf.reduce_sum(
            the_loss * loss_mask * self.loss_weights, axis=1
        )
        return tf.reduce_mean(sum_per_batch_element)

    def observe(self, samples, gt_resource_id):
        self.maybe_update_embedding()

    def regularization_losses(self):
        return self.fc_layer.losses

    def trainable_variables(self):
        return self.fc_layer.trainable_variables

    def save(self, path):
        with open(path + "_hc.pkl", "wb") as target:
            pickle.dump(self.fc_layer.get_weights(), target)

        with open(path + "_uidtodim.pkl", "wb") as target:
            pickle.dump((self.uid_to_dimension,), target)

    def restore(self, path):
        with open(path + "_hc.pkl", "rb") as target:
            new_weights = pickle.load(target)
            has_weights = False
            try:
                has_weights = len(self.fc_layer.get_weights()) == 2
            except Exception:
                pass

            if not has_weights:
                self.fc_layer.build([None, new_weights[0].shape[0]])

            self.fc_layer.set_weights(new_weights)

        with open(path + "_uidtodim.pkl", "rb") as target:
            (self.uid_to_dimension,) = pickle.load(target)

        self.update_embedding()



from chia.components.evaluators.evaluator import Evaluator


class TopKAccuracyEvaluator(Evaluator):
    def __init__(self, kb, kmax=5):
        self.kmax = kmax
        self.kb = kb
        self.reset()

    def reset(self):
        self.sample_count = 0
        self.correct_count = [0] * self.kmax

    def update(self, samples, gt_resource_id, prediction_resource_id):
        for sample in iter(samples):
            if sample.has_resource(prediction_resource_id + "_dist"):
                self.sample_count += 1
                gt = sample.get_resource(gt_resource_id)
                pred = list(
                    sorted(
                        sample.get_resource(prediction_resource_id + "_dist"),
                        key=lambda x: x[1],
                        reverse=True,
                    )
                )[: self.kmax]
                pred = [x[0] for x in pred]

                for k in range(self.kmax):
                    if gt in pred[: k + 1]:
                        self.correct_count[k] += 1

    def result(self):
        if self.sample_count > 0:
            return {
                "accuracy_at_k": [
                    float(correct_count) / float(self.sample_count)
                    for correct_count in self.correct_count
                ]
            }
        else:
            return dict()



import abc
import random
import time

import numpy as np
import tensorflow as tf

from chia import components, instrumentation
from chia.components.base_models.keras import keras_featureextractor, keras_preprocessor
from chia.helpers import batches_from_pair, make_generator_faster


class KerasTrainer(abc.ABC):
    @abc.abstractmethod
    def observe_inner(self, samples, gt_resource_id, progress_callback=None):
        pass

    @abc.abstractmethod
    def rehearse(self, steps, progress_callback=None):
        pass


class KerasFastSingleShotTrainer(KerasTrainer, instrumentation.Observable):
    def __init__(
        self,
        feature_extractor: keras_featureextractor.KerasFeatureExtractor,
        preprocessor: keras_preprocessor.KerasPreprocessor,
        classifier,
        learning_rate_schedule,
        optimizer,
        batch_size,
        inner_steps,
        sequential_training_batches=1,
    ):
        instrumentation.Observable.__init__(self)

        # Feature Extractor
        self.feature_extractor_new = feature_extractor
        self.preprocessor = preprocessor

        # Classifier
        self.classifier = classifier

        # Learning rate schedule
        self.learning_rate_schedule = learning_rate_schedule

        # Optimization
        self.optimizer = optimizer
        self.batch_size = batch_size
        self.sequential_training_batches = sequential_training_batches
        self._inner_steps = inner_steps

        # State here
        self.current_step = 0
        self._already_observed = False

    def perform_single_gradient_step(self, batch_elements_X, batch_elements_y):
        if self.feature_extractor_new.trainable:
            total_trainable_variables = (
                self.feature_extractor_new.feature_extractor.trainable_variables
                + self.classifier.trainable_variables()
            )
        else:
            total_trainable_variables = self.classifier.trainable_variables()

        # Forward step
        inner_bs = self.batch_size
        acc_gradients = None
        acc_hc_loss = 0
        inner_batch_count = 0

        for inner_batch_X, inner_batch_y in batches_from_pair(
            batch_elements_X, batch_elements_y, inner_bs
        ):
            # Build batch
            batch_X = self.preprocessor.preprocess_image_batch(
                np.stack(inner_batch_X, axis=0), is_training=True
            )
            batch_y = inner_batch_y

            # No numpy stacking here, these could be
            # strings or something else (concept uids)

            with tf.GradientTape() as tape:
                feature_batch = self.feature_extractor_new.feature_extractor(
                    batch_X, training=self.feature_extractor_new.trainable
                )
                hc_loss = self.classifier.loss(
                    feature_batch, batch_y, self.current_step
                )

                if self.feature_extractor_new.trainable:
                    reg_loss = sum(
                        self.feature_extractor_new.feature_extractor.losses
                        + self.classifier.regularization_losses()
                    )
                else:
                    reg_loss = self.classifier.regularization_losses()

                total_loss = hc_loss + reg_loss

            # Backward step
            gradients = tape.gradient(total_loss, total_trainable_variables)
            if acc_gradients is None:
                acc_gradients = gradients
            else:
                acc_gradients = [
                    acc_gradient + new_gradient
                    for acc_gradient, new_gradient in zip(acc_gradients, gradients)
                ]

            acc_hc_loss += hc_loss
            inner_batch_count += 1

        # Optimize
        self.optimizer.learning_rate = self.learning_rate_schedule(self.current_step)
        self.optimizer.apply_gradients(
            zip(
                [
                    acc_gradient / float(inner_batch_count)
                    for acc_gradient in acc_gradients
                ],
                total_trainable_variables,
            )
        )

        self.current_step += 1
        return acc_hc_loss / float(inner_batch_count)

    def observe_inner(self, samples, gt_resource_id, progress_callback=None):
        assert not self._already_observed, "This model can not learn continually"
        assert len(samples) > 0

        total_bs = self.batch_size * self.sequential_training_batches

        def my_gen():
            for inner_step in range(self._inner_steps):
                batch_samples = random.choices(samples, k=total_bs)
                batch_elements_y = []
                batch_elements_X = []
                for sample in batch_samples:
                    batch_elements_X.append(_get_input_img_np(sample))
                    batch_elements_y.append(sample.get_resource(gt_resource_id))

                yield inner_step, (batch_elements_X, batch_elements_y)

        if progress_callback is not None:
            progress_callback(0.0)

        hc_loss_running = 0.0
        time_per_step_running = 0.0
        hc_loss_factor = 0.0
        last_step_end_time = time.time()
        for inner_step, (X, y) in make_generator_faster(
            my_gen, "threading", observable=self, max_buffer_size=20
        ):
            if progress_callback is not None:
                progress_callback(inner_step / float(self._inner_steps))

            hc_loss = self.perform_single_gradient_step(X, y)

            hc_loss_running += hc_loss.numpy()
            hc_loss_factor += 1.0

            step_end_time = time.time()
            time_per_step_running += step_end_time - last_step_end_time

            if self.current_step % 10 == 9:
                self.report_metric(
                    "loss_ravg", hc_loss_running / hc_loss_factor, self.current_step
                )
                self.report_metric(
                    "time_per_step",
                    time_per_step_running / hc_loss_factor,
                    self.current_step,
                )
                hc_loss_running = 0.0
                hc_loss_factor = 0.0
                time_per_step_running = 0.0

            last_step_end_time = step_end_time

        if progress_callback is not None:
            progress_callback(1.0)

    def rehearse(self, steps, progress_callback=None):
        raise ValueError("Cannot learn continually!")


def _get_input_img_np(sample):
    return sample.get_resource("input_img_np")


class KerasTrainerFactory(components.Factory):
    name_to_class_mapping = {"fast_single_shot": KerasFastSingleShotTrainer}
    default_section = "keras_trainer"



import gzip
import json
import logging
import pathlib

from chia.instrumentation.message import (
    ConfigMessage,
    LogMessage,
    Message,
    MetricMessage,
    ResultMessage,
    ShutdownMessage,
)
from chia.instrumentation.observers.observer import Observer


class JSONObserver(Observer):
    message_filter = (ConfigMessage, ResultMessage, MetricMessage, ShutdownMessage)
    log_message_filter = (LogMessage,)

    def __init__(
        self,
        experiment_metadata=None,
        environment_info=None,
        path_pattern="results/%s/%s.json.gz",
        log_level=logging.WARNING,
        compress=True,
    ):
        # Deal with potential nonexistant metadata
        if experiment_metadata is not None:
            experiment_name = experiment_metadata.name
            experiment_run_id = experiment_metadata.run_id
        else:
            experiment_name = "no_name"
            experiment_run_id = "no_run_id"

        # Try using the path pattern
        try:
            self.path = path_pattern % (experiment_name, experiment_run_id,)
        except TypeError:
            self.path = path_pattern

        self.log_level = log_level
        self.compress = compress

        # Make sure the directory exists
        parent_path = pathlib.Path(self.path).parent
        if not parent_path.exists():
            parent_path.mkdir(parents=True)

        # Try writing to path to fail early
        with open(self.path, "w") as f:
            f.write("{}")

        self.valid_messages = {
            filter_element.__name__: []
            for filter_element in list(self.message_filter)
            + list(self.log_message_filter)
        }

        self.valid_messages["metadata"] = self._object_to_dict(experiment_metadata)
        self.valid_messages["environment"] = self._object_to_dict(environment_info)

    def update(self, message: Message):
        if any(
            isinstance(message, filter_element)
            for filter_element in self.message_filter
        ):
            message_dict = self._object_to_dict(message)
            self.valid_messages[message.__class__.__name__] += [message_dict]
        elif any(
            isinstance(message, filter_element)
            for filter_element in self.log_message_filter
        ):
            message_log_level = message.level
            if message_log_level >= self.log_level:
                message_dict = self._object_to_dict(message)
                self.valid_messages[message.__class__.__name__] += [message_dict]

        if isinstance(message, ShutdownMessage):
            self.write_json()

    def write_json(self):
        if self.compress:
            with gzip.open(self.path, "wt", encoding="ascii") as f:
                json.dump(obj=self.valid_messages, fp=f)
        else:
            with open(self.path, "w") as f:
                json.dump(obj=self.valid_messages, fp=f, indent=2)

    @staticmethod
    def _object_to_dict(obj):
        if hasattr(obj, "__dict__"):
            ret_dict = dict()
            for key, value in obj.__dict__.items():
                try:
                    json.dumps({key: value})
                    ret_dict[key] = value
                except (ValueError, TypeError) as err:
                    ret_dict[key] = f"Could not serialize :( Error: {str(err)}"

            return ret_dict
        elif hasattr(obj, "_asdict"):
            return obj._asdict()
        else:
            return {"string_repr": str(obj)}
